class Url:
    @staticmethod
    def oauth():
        return "https://auth.tdameritrade.com/auth"

    @staticmethod
    def quote():
        return "https://api.tdameritrade.com/v1/marketdata/{}/quotes"



from broker.paper_trade import Paper
from broker.robinhood_api import Robinhood


def main():
    c = Robinhood()
    # c.prompt_login()
    paper_trader = Paper(client=c, cash=100000.0)
    return paper_trader


client = None
if __name__ == "__main__":
    Robinhood.INSECURE = True
    client = main()



import datetime
from time import sleep


class BookAnalyzer:
    def __init__(self):
        self.history = []
        self.time_ratio = []
    
    
    def load_data(self, filename):
        self.history = []
        self.time_ratio = []
        with open(filename, 'r') as f:
            for line in f:
                if line == '':
                    break
                line = line.replace('\n', '')
                arr = line.split(' ')
                bids = []
                asks = []
                count = 0
                for i in range(1, len(arr)):
                    pair = arr[i][1:len(arr[i]) - 1]
                    pair = pair.split(',')
                    if count < 10:
                        bids.append((float(pair[0]), int(pair[1])))
                        count += 1
                    else:
                        asks.append((float(pair[0]), int(pair[1])))
                # Dependent on decision to print bids in order, or from price to 10 ticks out
                bids.reverse()
                self.history.append((arr[0], bids, asks))
                # [0] is time, [1] to [10] are (bid price, bid size), [11] to [21] are (ask price, ask size)
            f.close()
        
    def set_ratio(self):
        for row in self.history:
            bid_orders = 0
            ask_orders = 0
            for bid in row[1]:
                bid_orders += bid[1]
            for ask in row[2]:
                ask_orders += ask[1]
            self.time_ratio.append((row[0], bid_orders, ask_orders))
            if ask_orders == 0:
                print(row[0] + "ASK ORDERS 0")
                return
            print(str(bid_orders) + "\t" + str(-1 * ask_orders))
            
    def assume_ratio_changes(self):
        prev_bid = self.history[0][1][0]
        prev_ask = self.history[0][2][0]
        output = [[0,0],[0,0]]
        for i in range(1, len(self.history)):
            if self.history[i][1][0] != prev_bid:
                is_up = prev_bid < self.history[i][1][0]
                
                prev_bid = self.history[i][1][0]
                prev_ask = self.history[i][2][0]
                # print(self.history[i][0] + " " + str(is_up))
                
                ratio_increase = self.time_ratio[i - 1][1] / float(self.time_ratio[i - 1][2]) < self.time_ratio[i][1] / float(self.time_ratio[i][2])
                a = 0 if is_up else 1
                b = 0 if ratio_increase else 1
                output[a][b] += 1
                # print(self.history[i][0] + "\t" + str(is_up) + "\t" + str(self.time_ratio[i - 1][1] / float(self.time_ratio[i - 1][2])) + "\t->\t" + str(self.time_ratio[i][1] / float(self.time_ratio[i][2])))
                
            elif self.history[i][2][0] != prev_ask:
                is_up = prev_ask < self.history[i][2][0]
                prev_bid = self.history[i][1][0]
                prev_ask = self.history[i][2][0]
                # print(self.history[i][0] + " " + str(is_up))
                
                ratio_increase = self.time_ratio[i - 1][1] / float(self.time_ratio[i - 1][2]) < self.time_ratio[i][1] / float(self.time_ratio[i][2])
                a = 0 if is_up else 1
                b = 0 if ratio_increase else 1
                output[a][b] += 1
                # print(self.history[i][0] + "\t" + str(is_up) + "\t" + str(self.time_ratio[i - 1][1] / float(self.time_ratio[i - 1][2])) + "\t->\t" + str(self.time_ratio[i][1] / float(self.time_ratio[i][2])))
                
            else:
                # print(self.history[i][0] + " ")
                pass
        # print(output)

    def print_spread(self):
        for row in self.history:
            print(str(row[1][0][0]) + "\t" + str(row[2][0][0]))


# Used to generate .book files
class BookLogger:
    def __init__(self, client, reps=10, delay=1):
        self.client = client
        if not self.client.logged_in:
            raise Exception('Need to login to client')
        # how many rows there will be
        self.reps = reps
        # how frequent to query the book in seconds
        self.delay = delay
    
    # Creates a file named 'aapl.book' or if id is given it will be (id='_3') 'aapl_3.book'
    # Format of file is:
    # [time] [bid -1 (price, size)] ... [bid -9 (price, size)] [bid -10 (price, size)] [ask +1 (price, size)] ... [ask +9 (price, size)] [ask +10 (price, size)]
    def collect(self, ticker, id=''):
        f = open(ticker + id + '.book', 'w')
        
        count = 0
        while count < self.reps:
            obj = self.client.get_book(ticker)
            bids = [(float(b['price']['amount']), int(b['quantity'])) for b in obj['bids'][:10]]
            asks = [(float(b['price']['amount']), int(b['quantity'])) for b in obj['asks'][:10]]
            time = datetime.datetime.now().strftime("%H:%M:%S")
            row = [str(time)]
            row += bids + asks
            
            result = row[0]
            for r in row[1:]:
                result += " {},{}".format(r[0], r[1])
            
            result += '\n'
            f.write(result)
            count += 1
            if count >= self.reps:
                break
            sleep(self.delay)
        
        f.close()


class SlimBook(BookLogger):
    def collect(self, ticker, id=''):
        f = open(ticker + id + '.book', 'w')
        
        count = 0
        while count < self.reps:
            obj = self.client.get_quote(ticker)
            time = datetime.datetime.now().strftime("%H:%M:%S")
            
            result = time + " {},{}".format(obj.bid_price, obj.bid_size) + " {},{}".format(obj.ask_price, obj.ask_size) + '\n'
            
            f.write(result)
            count += 1
            if count >= self.reps:
                break
            sleep(self.delay)
        
        f.close()

'''        
if __name__ == "__main__":
    ba = BookAnalyzer()
    ba.load_data("uber.book")
    ba.set_ratio()
    # ba.assume_ratio_changes()
'''
    


import getpass
import random
import requests


class Configuration:
    def __init__(self):
        self.username = None
        self.password = None
        self.device_id = None


class Quote:
    """
    Used as a base class for quote type.
    Implementations can reimplement the constructor to parse quote objects
    """
    def __init__(self):
        # Last traded price
        self.price = 0.0
        self.bid_price = 0.0
        self.bid_size = 0
        self.ask_price = 0.0
        self.ask_size = 0


class ApiBase:
    DEBUG = False
    INSECURE = False
    VERSION = "1.0"

    @staticmethod
    def gen_client():
        string = ""
        for i in range(0, 8):
            string += "{:01x}".format(random.randint(0, 15))
        string += "-"
        for i in range(0, 4):
            string += "{:01x}".format(random.randint(0, 15))
        string += "-"
        for i in range(0, 4):
            string += "{:01x}".format(random.randint(0, 15))
        string += "-"
        for i in range(0, 4):
            string += "{:01x}".format(random.randint(0, 15))
        string += "-"
        for i in range(0, 12):
            string += "{:01x}".format(random.randint(0, 15))
        return string
    
    def __init__(self):
        self.session = requests.Session()
        self.logged_in = False
        self.device_id = ApiBase.gen_client()
    
    def prompt_login(self):
        user = input("Username: ")
        pwd = getpass.getpass()
        self.login(user, pwd)
    
    def login(self, username, password):
        pass

    def insecure_login(self):
        pass
    
    def account_info(self):
        pass
    
    def logout(self):
        pass
    
    # Market buy
    def buy(self, symbol, quantity, extended=False):
        pass
    
    def limit_buy(self, symbol, price, quantity, extended=False, cancel=None):
        pass
    
    # Market sell
    def sell(self, symbol, quantity, extended=False):
        pass
    
    def limit_sell(self, symbol, price, quantity, extended=False, cancel=None):
        pass
    
    def get_quote(self, symbol):
        pass
    
    # Portfolio of stocks
    def get_positions(self):
        pass



class Url:
    api = 'https://api.robinhood.com'

    def accounts():
        return Url.api + '/accounts/'
    
    def book(s_id):
        return Url.api + '/marketdata/pricebook/snapshots/' + s_id + '/'
    
    def challenge(c_id):
        return Url.api + "/challenge/" + c_id + "/respond/"
    
    def historical(symbol, interval, span):
        return Url.api + "/marketdata/historicals/" + symbol + "/?bounds=regular&interval=" + interval + "&span=" + span
    
    def instruments(symbol):
        return Url.api + "/instruments/?symbol=" + symbol
    
    def login():
        return Url.api + '/oauth2/token/'
        
    def login_page():
        return "https://robinhood.com/login"
    
    def logout():
        return Url.api + '/oauth2/revoke_token/'
    
    def order():
        return Url.api + '/orders/'
    
    def positions():
        return Url.api + "/positions/?nonzero=true"
    
    def quote(s_id):
        return Url.api + "/marketdata/quotes/" + s_id + "/?include_inactive=true"


class Robinhood:
    def __init__(self, client):
        self.client = client
    
    def login(self):
        self.client.prompt_login()
    
    '''
    To maintain a session instead of entering SMS every time create a configuration file as detailed:
    config.py:
    Then make sure Client.INSECURE = True
    '''
    def insecure_login(self):
        self.client.insecure_login()
    
    def portfolio(self):
        pos = self.client.get_positions()
        if pos is None:
            return
        for p in pos:
            symbol = self.client.get_symbol_from_instrument(p['instrument'])
            quantity = int(float(p['quantity']))
            price = float(p['average_buy_price'])
            current_price = self.client.get_price(symbol).price
            print("{}\t{}\t{}\t{}".format(symbol, quantity, price, current_price))
    
    def buy(self, symbol, quantity, extended=False, cancel=None):
        self.client.buy(symbol, quantity, extended, cancel)
    
    def limit_buy(self, symbol, price, quantity, extened=False, cancel=None):
        self.client.limit_buy(symbol, price, quantity, extened, cancel)
        
    def sell(self, symbol, quantity, extended=False, cancel=None):
        self.client.sell(symbol, quantity, extended, cancel)
    
    def limit_sell(self, symbol, price, quantity, extended=False, cancel=None):
        self.client.limit_sell(symbol, price, quantity, extended, cancel)
    
    def cancel_order(self):
        length = len(self.client.pending_orders)
        if length is 1:
            self.client.cancel_order(0)
        elif length is 0:
            print('No orders to cancel')
            return
        else:
            for i in range(0, len(self.client.pending_orders)):
                # SYMBOL\tquantity\tprice
                print(str(i) + ": " + self.client.pending_orders[i]["side"] \
                      + "\t" + self.client.get_symbol_from_instrument(self.client.pending_orders[i]["instrument"]) \
                      + "\t" + str(int(float(self.client.pending_orders[i]["quantity"]))) \
                      + "\t" + str(float(self.client.pending_orders[i]["price"])))
            num = int(input("Enter order number to cancel: "))
            if num >= 0 and num < length:
                self.client.cancel_order(num)
            else:
                print("Forbidden order number: " + str(num))
            
    def get_total(self):
        acct = self.client.account_info()
        return float(acct['results'][0]['cash']) + float(acct['results'][0]['unsettled_funds']) + float(acct['results'][0]['uncleared_deposits'])
    
    def logout(self):
        self.client.logout()



from .ameritrade_url import Url
from .api_base import ApiBase, Configuration, Quote
from ssl import SSLError
import time
import requests
import pickle
import json


class AQuote(Quote):
    def __init__(self, obj):
        super().__init__()
        self.price = obj['lastPrice']
        if 'askPrice' in obj:
            self.ask_price = obj['askPrice']
        if 'askSize' in obj:
            self.ask_size = obj['askSize']
        if 'bidPrice' in obj:
            self.bid_price = obj['bidPrice']
        if 'bidSize' in obj:
            self.bid_size = obj['bidSize']


class Ameritrade(ApiBase):
    def __init__(self, client_id="QOPIXDKLBHG6PXRAHNFCR1LC58ZMMFHX"):
        super().__init__()
        self.session.headers = {
            'Connection': 'keep-alive',
            'DNT': '1'
        }
        self.client_id = client_id + "@AMER.OAUTHAP"

    '''
    Login to TD Ameritrade. Ameritrade wants to force this to be a web browser implementation
    by requiring a redirect url which will post the code to.
    In our case we can force it to be a RESTful client and store the code from response. (need to test this method)
    Sequence is:
        post -> https://auth.tdameritrade.com/auth -> save code
        post -> https://api.tdameritrade.com/v1/oauth2/token
    '''
    def login(self, username, password):
        params = {
            "response_type": "code",
            "redirect_uri": "http://localhost:80",
            "client_id": self.client_id
        }
        headers = {"Content-Type": "x-www-form-urlencoded"}
        payload = {
            "_csfr": self.device_id,
            "s_scope": "PlaceTrades+AccountAccess+MoveMoney",
            "s_client-name": self.client_id,
            "fp_fp2DeviceId": "8b9d59b77e16496f94bfda8be55b060a",
            "fp_browser": "mozilla/5.0+(windows+nt+10.0;+win64;+x64;+rv:73.0)+gecko/20100101+firefox/73.0|5.0+(Windows)|Win32",
            "fp_screen": "24|1536|864|824",
            "fp_timezone": int(-time.timezone / 60 / 60),
            "fp_language": "en-US|syslang=|userlang=",
            "fp_java": 0,
            "fp_cookie": 1,
            "fp_cfp": '{"navigator":{"properties":{"doNotTrack":"unspecified","maxTouchPoints":"0","oscpu":"Windows+NT+10.0;+Win64;+x64","vendor":"","vendorSub":"","productSub":"20100101","cookieEnabled":"true","buildID":"20181001000000","webdriver":"false","hardwareConcurrency":"12","appCodeName":"Mozilla","appName":"Netscape","appVersion":"5.0+(Windows)","platform":"Win32","userAgent":"Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64;+rv:73.0)+Gecko/20100101+Firefox/73.0","product":"Gecko","language":"en-US","onLine":"true"},"propertiesMD5":"551c644b8bee69e6c9c1371351fc29bb","navorderMD5":"d6e0b898d210c876d0c4e4eced591cd6","navmethods":"vibrate|javaEnabled|getGamepads|getVRDisplays|mozGetUserMedia|sendBeacon|requestMediaKeySystemAccess|registerProtocolHandler|taintEnabled","navmethMD5":"98288d814fb8f7a931b503c9a99e8d45"},"screen":{"properties":{"availWidth":"1536","availHeight":"824","width":"1536","height":"864","colorDepth":"24","pixelDepth":"24","top":"0","left":"0","availTop":"0","availLeft":"0","mozOrientation":"landscape-primary","onmozorientationchange":"NULL"},"propertiesMD5":"8ed1699940e8f38cda24729715b459cb","tamper":false},"timezone":"-6|-5","canvas":"f3b84e719180c3c5b37da8ac9a4b3e1e","java":false,"localstorage":true,"sessionstorage":true,"indexedDB":true,"plugins":{"names":"","namesMD5":"d41d8cd98f00b204e9800998ecf8427e"},"fonts":"2e8ad97446bda38227f4904a8753f7fc","mathroutines":"11013.232920103324|-1.4214488238747245","md5":"118757b6c1fd15a5648814139690299d","latency":164,"form":{"fields":"su_username|su_password|rememberuserid|authorize|reject"},"clkhz":null}',
            "su_username": username,
            "su_password": password,
            "authorize": "Log+in",
            "signed": "24879f023abdeabba9d2c75eb8198e78a7c7b428"
        }
        if Ameritrade.DEBUG:
            req = requests.Request('POST', Url.oauth(), params=params, headers=headers, data=payload)
            pretty_print_post(req)

        try:
            response = self.session.post(Url.oauth(), params=params, headers=headers, data=payload)
            pretty_print_post(response)
        except SSLError:
            print("SSL exception, login failed.")
            return
        except Exception:
            print("Login failed.")
            return
        print("Login successful.")
        self.logged_in = True
        # When we successfully we login we can save the user/pass insecurely. Please do not share configuration.pkl file
        if Ameritrade.INSECURE:
            config = Configuration()
            config.username = username
            config.password = password
            config.device_id = self.device_id
            with open('configuration.pkl', 'wb') as f:
                pickle.dump(config, f, pickle.HIGHEST_PROTOCOL)

    # Login with already logged in credentials. Do not distribute configuration.pkl
    def insecure_login(self):
        if not Ameritrade.INSECURE:
            return
        try:
            with open('configuration.pkl', 'rb') as f:
                config = pickle.load(f)
        except FileNotFoundError:
            self.prompt_login()
            return

        self.device_id = config.device_id
        self.login(config.username, config.password)

    # https://developer.tdameritrade.com/quotes/apis/get/marketdata/%7Bsymbol%7D/quotes
    def get_quote(self, symbol):
        url = Url.quote().format(symbol)
        response = self.session.get(url)
        obj = json.loads(response.text)
        quote = AQuote(obj)
        return quote


def pretty_print_post(req):
    req = req.prepare()
    """
    At this point it is completely built and ready
    to be fired; it is "prepared".

    However pay attention at the formatting used in 
    this function because it is programmed to be pretty 
    printed and may differ from the actual request.
    """
    print('{}\n{}\r\n{}\r\n\r\n{}'.format(
        '-----------START-----------',
        req.method + ' ' + req.url,
        '\r\n'.join('{}: {}'.format(k, v) for k, v in req.headers.items()),
        req.body,
        ))



from .api_base import ApiBase
import datetime
import json


class Paper(ApiBase):
    def __init__(self, client=None, cash=0.0, portfolio=None):
        """
        client: either Robinhood or TD Ameritrade account. The account must be logged in; used for getting price quotes.
        cash: available cash, default is 0.0
        portfolio: structured as follows:
            [
                {
                    "symbol": string,
                    "price": float,
                    "quantity": int,
                    "purchase_date": string ("MM/DD/YYYY HH:MM:SS.sss Z")
                }
            ]
        """
        self._account = {
            "cash": cash,
            "portfolio": portfolio,
            "total": 0.0
        }
        # WARNING, DO NOT USE CLIENT TO ACTUALLY BUY/SELL
        self._client = client
        super().__init__()

    def _update_total(self):
        if self._account['portfolio'] is None:
            self._account['total'] = self._account['cash']
            return

        total = 0.0
        for symbol in self._account['portfolio']:
            for purchase in self._account['portfolio'][symbol]:
                total += self.get_quote(symbol).price * purchase['quantity']
        self._account['total'] = total + self._account['cash']

    def get_quote(self, symbol):
        return self._client.get_quote(symbol)

    def buy(self, symbol, quantity, extended=False):
        quote = self.get_quote(symbol)
        if quote is None:
            print("Cannot retrieve quote for {}".format(symbol))
            return
        total = quote.price * quantity
        if total > self._account['cash']:
            # format string for $
            print("Not enough funds in account to buy {} shares of {} at ${}".format(quantity, symbol, quote.price))
            return

        if self._account['portfolio'] is None:
            self._account['portfolio'] = {}
        if symbol not in self._account['portfolio']:
            self._account['portfolio'][symbol] = []
        now = datetime.datetime.now()
        self._account['portfolio'][symbol].append({
            "price": quote.price,
            "quantity": quantity,
            "purchase_date": now.strftime("%d/%m/%Y %H:%M:%S.%f")[:-3] + now.strftime(" %Z")
        })
        self._account['cash'] -= total

    def sell(self, symbol, quantity, extended=False):
        if symbol not in self._account['portfolio']:
            print("short selling not implemented, no margin.")
            return
        quote = self.get_quote(symbol)
        if quote is None:
            print("Cannot retrieve quote for {}".format(symbol))
            return
        total = quote.price * quantity

        if self._account['portfolio'] is None:
            self._account['portfolio'] = {}
        if symbol not in self._account['portfolio']:
            self._account['portfolio'][symbol] = []
        now = datetime.datetime.now()
        # For total portfolio tracking it records all purchasing/selling transactions
        self._account['portfolio'][symbol].append({
            "price": quote.price,
            "quantity": -quantity,
            "purchase_date": now.strftime("%d/%m/%Y %H:%M:%S.%f")[:-3] + now.strftime(" %Z")
        })
        self._account['cash'] += total

    def login(self, username, password):
        self._client.login(username, password)

    def insecure_login(self):
        self._client.insecure_login()

    def logout(self):
        self._client.logout()

    def save(self):
        # self._update_total()
        with open('portfolio.json', 'w') as f:
            json.dump(self._account, f)

    def load(self):
        with open('portfolio.json', 'r') as f:
            line = f.readline()
            self._account = json.loads(line)



import json
import pickle

from broker.api_base import ApiBase, Configuration, Quote
from broker.robinhood_url import Url


class RQuote(Quote):
    def __init__(self, obj):
        super().__init__()
        self.price = float(obj['last_trade_price'])
        self.bid_price = float(obj['bid_price'])
        self.bid_size = obj['bid_size']
        self.ask_price = float(obj['ask_price'])
        self.ask_size = obj['ask_size']
        self.prev_close = float(obj['adjusted_previous_close'])
        
    def __str__(self):
        return "{ " + "price: " + str(self.price) + ", bid_price: " + str(self.bid_price) + ", bid_size: "\
            + str(self.bid_size) + ", ask_price: " + str(self.ask_price) + ", ask_size: " + str(self.ask_size)\
            + ", prev_close: " + str(self.prev_close) + " }"


class Robinhood(ApiBase):
    def __init__(self):
        super().__init__()
        self.session.headers = {
            'Accept': '*/*',
            'Accept-Language': 'en-US,en;q=0.5',
            'Accept-Encoding': 'gzip, deflate, br',
            'X-Robinhood-API-Version': '1.275.0',
            'Connection': 'keep-alive',
            'DNT': '1',
            'TE': 'Trailers'
        }
        self.client_id = "c82SH0WZOsabOXGP2sxqcj34FxkvfnWRZBKlBjFS"
        self.refresh_token = None
        self.account = None
        self.instruments = {}
        self.stock_ids = {}
        self.symbols = {}
        self.pending_orders = []
        # print('constructed ' + Client.VERSION)
    
    def get_device_token(self):
        # clientId: "abcd-1938adf-192398192afasd-1239a"
        resp = self.session.get(Url.login_page())
        if Robinhood.DEBUG:
            print(resp.status_code)
    
    def sms_confirm(self, username, password):
        data = {
            "grant_type":"password",
            "scope":"internal",
            "client_id":self.client_id,
            "expires_in":86400,
            # Device token should be user-input
            "device_token":self.device_id,
            "username":username,
            "password":password,
            "challenge_type": "sms"
        }
        string = json.dumps(data)
        if Robinhood.DEBUG:
            print(Url.login())
            print(self.session.headers)
            print(string)
            print(len(string))
        
        resp = self.session.post(Url.login(), data=string, headers={'Content-Type': 'application/json', 'Content-Length': str(len(string))})
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        # when login fails, 400 error code is returned
        if resp.status_code == 200:
            return
        
        obj = json.loads(resp.text)
        c_id = obj['challenge']['id']
        sms = input("SMS Code: ")
        data_resp = {
            'response': sms
        }
        resp = self.session.post(
            Url.challenge(c_id),
            data=json.dumps(data_resp),
            headers={'Content-Type': 'application/json'})
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        # successful challenge accepted
        if resp.status_code != 200:
            # possibly throw an exception instead
            print('Wrong SMS code, login failed.')
        else:
            resp = self.session.post(
                Url.login(),
                data=json.dumps(data),
                headers={'Content-Type': 'application/json', 'X-ROBINHOOD-CHALLENGE-RESPONSE-ID': c_id})
            if Robinhood.DEBUG:
                Robinhood.log_response(resp)
            if resp.status_code != 200:
                # possibly throw exception instead
                print('login failed')
                return
            obj = json.loads(resp.text)
            self.refresh_token = obj['refresh_token']
            self.session.headers['Authorization'] = 'Bearer ' + obj['access_token']
            self.logged_in = True
            # make account request to get current info
            self.account = self.account_info()['results'][0]
    
    '''
    login: make login request, and then get account info (ignore if logged in already)
    username: "test@mail.com"
    password: "password"
    '''
    def login(self, username, password):
        if self.logged_in:
            return
        device_token = ""
        # self.get_device_token()
        self.sms_confirm(username, password)
        # INSECURE LOGIN, FILE SAVED LOCALLY. POTENTIALLY MALICIOUS APPLICATIONS CAN FIND THIS FILE ######
        # TODO ENCRYPT/DECRYPT CONFIGURATION FILE
        if Robinhood.INSECURE:
            config = Configuration()
            config.username = username
            config.password = password
            with open('configuration.pkl', 'wb') as f:
                pickle.dump(config, f, pickle.HIGHEST_PROTOCOL)
        
    '''
    Insecure login
    '''
    def insecure_login(self):
        if not Robinhood.INSECURE:
            return
        # INSECURE LOGIN, FILE SAVED LOCALLY. POTENTIALLY MALICIOUS APPLICATIONS CAN FIND THIS FILE ######
        config = None
        try:
            with open('configuration.pkl', 'rb') as f:
                config = pickle.load(f)
        except FileNotFoundError:
            self.prompt_login()
            return
        self.session.cookies['device_id'] = self.device_id
        data = {
            "grant_type": "password",
            "scope": "internal",
            "client_id": self.client_id,
            "expires_in": 86400,
            # Device token should be user-input
            "device_token": self.device_id,
            "username": config.username,
            "password": config.password
        }

        resp = self.session.post(Url.login(), data=json.dumps(data), headers={'Content-Type': 'application/json'})
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        if resp.status_code != 200:
            # possibly throw exception instead
            print('login failed')
            return
        obj = json.loads(resp.text)
        self.refresh_token = obj['refresh_token']
        self.session.headers['Authorization'] = 'Bearer ' + obj['access_token']
        self.logged_in = True
        # make account request to get current info
        self.account = self.account_info()['results'][0]

    '''
    account_info: requires to be logged in
    return: json object of /accounts/ request
    '''
    def account_info(self):
        if not self.logged_in:
            return None
        resp = self.session.get(Url.accounts())
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return json.loads(resp.text)

    '''
    logout: requires to be logged in, requests token to be removed
    TODO: which token is used???
    '''
    def logout(self):
        if not self.logged_in:
            return
        data = {
            "client_id":self.client_id,
            "token":self.refresh_token
        }
        resp = self.session.post(Url.logout(), data=json.dumps(data), headers={'Content-Type': 'application/json'})
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        self.logged_in = False

    '''
    limit_buy: buy a stock at a limit price (usually lower than current price)
    symbol: "ABC"
    price: "10.01"
    quantity: "10"
    extended: True
    cancel: "gtc"
    '''
    def limit_buy(self, symbol, price, quantity, extended=False, cancel=None):
        if not self.logged_in:
            self.prompt_login()
        # This also will update stock_ids
        symbol = symbol.upper()
        instrument = self.get_instrument(symbol)
        if cancel is None:
            cancel = "gfd"
        data = {
            "time_in_force":cancel,
            "price":price,
            "quantity":quantity,
            "side":"buy",
            "trigger":"immediate",
            "type":"limit",
            "account":self.account['url'],
            "instrument":instrument,
            "symbol":symbol,
            # "ref_id":"",
            "extended_hours":extended
        }
        if Robinhood.DEBUG:
            print(data)
        
        resp = self.session.post(Url.order(), data=json.dumps(data), headers={'Content-Type': 'application/json'})
        obj = json.loads(resp.text)
        self.pending_orders.append(obj)
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return obj
    
    def buy(self, symbol, quantity, extended=False, cancel=None):
        if not self.logged_in:
            self.prompt_login()
        # This also updates stock_ids
        symbol = symbol.upper()
        instrument = self.get_instrument(symbol)
        if cancel is None:
            cancel = "gfd"
        data = {
            "time_in_force":cancel,
            "quantity":quantity,
            "side":"buy",
            "trigger":"immediate",
            "type":"market",
            "account":self.account['url'],
            "instrument":instrument,
            "symbol":symbol,
            # "ref_id":"",
            "extended_hours":extended
        }
        if Robinhood.DEBUG:
            print(data)
        
        resp = self.session.post(Url.order(), data=json.dumps(data), headers={'Content-Type': 'application/json'})
        obj = json.loads(resp.text)
        self.pending_orders.append(obj)
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return obj
    
    def limit_sell(self, symbol, price, quantity, extended=False, cancel=None):
        if not self.logged_in:
            self.prompt_login()
        # This also will update stock_ids
        symbol = symbol.upper()
        instrument = self.get_instrument(symbol)
        if cancel is None:
            cancel = "gfd"
        data = {
            "time_in_force":cancel,
            "price":price,
            "quantity":quantity,
            "side":"sell",
            "trigger":"immediate",
            "type":"limit",
            "account":self.account['url'],
            "instrument":instrument,
            "symbol":symbol,
            # "ref_id":"",
            "extended_hours":extended
        }
        if Robinhood.DEBUG:
            print(data)
        
        resp = self.session.post(Url.order(), data=json.dumps(data), headers={'Content-Type': 'application/json'})
        obj = json.loads(resp.text)
        self.pending_orders.append(obj)
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return obj
    
    def sell(self, symbol, quantity, extended=False, cancel=None):
        if not self.logged_in:
            self.prompt_login()
        # This also updates stock_ids
        symbol = symbol.upper()
        instrument = self.get_instrument(symbol)
        if cancel is None:
            cancel = "gfd"
        
        data = {
            "time_in_force":cancel,
            "quantity":quantity,
            "side":"sell",
            "trigger":"immediate",
            "type":"market",
            "account":self.account['url'],
            "instrument":instrument,
            "symbol":symbol,
            # "ref_id":"",
            "extended_hours":extended
        }
        if Robinhood.DEBUG:
            print(data)
        
        resp = self.session.post(Url.order(), data=json.dumps(data), headers={'Content-Type': 'application/json'})
        obj = json.loads(resp.text)
        self.pending_orders.append(obj)
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return obj
    
    def get_instrument(self, symbol):
        if symbol in self.instruments.keys():
            return self.instruments[symbol]
        else:
            resp = self.session.get(Url.instruments(symbol=symbol))
            if Robinhood.DEBUG:
                Robinhood.log_response(resp)
            obj = json.loads(resp.text)
            url = obj['results'][0]['url']
            self.instruments[symbol] = url
            self.stock_ids[symbol] = obj['results'][0]['id']
            self.symbols[obj['results'][0]['id']] = symbol
            return url
    
    '''
    get_quote: returns Quote
    symbol: Tradable stock symbol name
    {
      "ask_price": "48.190000",
      "ask_size": 2300,
      "bid_price": "48.180000",
      "bid_size": 2900,
      "last_trade_price": "48.309300",
      "last_extended_hours_trade_price": null,
      "previous_close": "46.850000",
      "adjusted_previous_close": "46.850000",
      "previous_close_date": "2019-06-25",
      "symbol": "INTC",
      "trading_halted": false,
      "has_traded": true,
      "last_trade_price_source": "nls",
      "updated_at": "2019-06-26T17:52:07Z",
      "instrument": "https:\/\/api.robinhood.com\/instruments\/ad059c69-0c1c-4c6b-8322-f53f1bbd69d4\/"
    }
    '''
    def get_quote(self, symbol):
        symbol = symbol.upper()
        self.get_instrument(symbol)
        # stock_ids is updated
        s_id = self.stock_ids[symbol]
        resp = self.session.get(Url.quote(s_id))
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return RQuote(json.loads(resp.text))
    
    # https://api.robinhood.com/marketdata/historicals/SPY/?bounds=trading&interval=5minute&span=day
    # symbol = "SPY"
    # interval = 5minute, 10minute, day, week
    # span = day, month, year, 5year
    '''
    {
        "quote":"https://api.robinhood.com/quotes/8f92e76f-1e0e-4478-8580-16a6ffcfaef5/",
        "symbol":"SPY",
        "interval":"5minute",
        "span":"day",
        "bounds":"trading",
        "previous_close_price":"292.620000",
        "previous_close_time":"2019-08-02T20:00:00Z",
        "open_price":"288.200000",
        "open_time":"2019-08-05T13:00:00Z",
        "instrument":"https://api.robinhood.com/instruments/8f92e76f-1e0e-4478-8580-16a6ffcfaef5/",
        "historicals":
        [{
            "begins_at":"2019-08-05T13:00:00Z",
            "open_price":"288.200000",
            "close_price":"288.190000",
            "high_price":"288.240000",
            "low_price":"287.970000",
            "volume":74159,
            "session":"pre",
            "interpolated":false
        }]
    }
    '''
    def get_historical(self, symbol, interval, span):
        symbol = symbol.upper()
        resp = self.session.get(Url.historical(symbol, interval, span))
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return json.loads(resp.text)
    
    '''
    'bids' {
        [{'side': 'bid', 'price': {'amount': '198.000000', 'currency_code': 'USD'}, 'quantity': 500}]
    },
    'asks' {
        [{'side': 'ask', 'price': {'amount': '198.010000', 'currency_code': 'USD'}, 'quantity': 500}]
    }
    '''
    def get_book(self, symbol):
        if not self.logged_in:
            return None
        symbol = symbol.upper()
        self.get_instrument(symbol)
        resp = self.session.get(Url.book(self.stock_ids[symbol]))
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        return json.loads(resp.text)
    
    '''
    {
      "shares_held_for_stock_grants": "0.0000",
      "account": "https:\/\/api.robinhood.com\/accounts\/5RX37639\/",
      "pending_average_buy_price": "160.7200",
      "shares_held_for_options_events": "0.0000",
      "intraday_average_buy_price": "0.0000",
      "url": "https:\/\/api.robinhood.com\/positions\/5RX37639\/9c53326c-d07e-4b82-82d2-b108ec5d9530\/",
      "shares_held_for_options_collateral": "0.0000",
      "created_at": "2018-04-03T17:15:10.913191Z",
      "updated_at": "2018-04-03T17:18:47.026106Z",
      "shares_held_for_buys": "0.0000",
      "average_buy_price": "160.7200",
      "instrument": "https:\/\/api.robinhood.com\/instruments\/9c53326c-d07e-4b82-82d2-b108ec5d9530\/",
      "intraday_quantity": "0.0000",
      "shares_held_for_sells": "0.0000",
      "shares_pending_from_options_events": "0.0000",
      "quantity": "1.0000"
    }
    '''
    def get_positions(self):
        if not self.logged_in:
            return None
        resp = self.session.get(Url.positions())
        return json.loads(resp.text)['results']
    
    def get_symbol_from_instrument(self, instrument):
        # 9c53326c-d07e-4b82-82d2-b108ec5d9530
        length = len(instrument)
        START_INDEX = 37
        END_INDEX = 1
        string = instrument[length - START_INDEX:length - END_INDEX]
        if string in self.symbols.keys():
            return self.symbols[string]
        else:
            resp = self.session.get(instrument)
            if Robinhood.DEBUG:
                Robinhood.log_response(resp)
            obj = json.loads(resp.text)
            self.symbols[string] = obj['symbol']
            self.instruments[obj['symbol']] = instrument
            self.stock_ids[obj['symbol']] = string
            return obj['symbol']
    
    def cancel_order(self, order_pos):
        if order_pos >= len(self.pending_orders):
            return
        
        resp = self.session.post(self.pending_orders[order_pos]['cancel'])
        if Robinhood.DEBUG:
            Robinhood.log_response(resp)
        del self.pending_orders[order_pos]

    @staticmethod
    def log_response(resp):
        print("--------START--------")
        print(resp.status_code)
        print(resp.reason)
        print(resp.headers)
        print(resp.text)
        print("---------END---------")
